from datetime import datetime
import logging
import os
from platformdirs import user_data_path
from tk3u8.constants import APP_NAME


def setup_logging(log_level: str | None) -> None:
    if not log_level:
        logging.basicConfig(level=logging.CRITICAL + 1)  # Avoid printing of log messages
        return

    log_directory = os.path.join(user_data_path(), APP_NAME, "logs")

    if not os.path.exists(log_directory):
        os.makedirs(log_directory, exist_ok=True)

    log_filename = f"logs-{datetime.now().strftime('%Y%m%d')}.log"
    log_file = os.path.join(log_directory, log_filename)

    file_handler = logging.FileHandler(log_file, encoding='utf-8')
    file_handler.setLevel(log_level)
    file_handler.setFormatter(logging.Formatter('%(asctime)s | %(levelname)s | %(name)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S%z'))

    logging.basicConfig(
        level=logging.DEBUG,
        handlers=[file_handler]
    )



import logging
import os
from typing import Optional
from platformdirs import user_data_path, user_downloads_path
import toml
from tk3u8.cli.console import console
from tk3u8.constants import APP_NAME, DEFAULT_CONFIG
from tk3u8.messages import messages

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class PathsHandler:
    """
    Handles the initialization and manages important file and directory
    paths for the application
    """

    def __init__(
            self,
            program_data_dir: Optional[str] = None,
            config_file_path: Optional[str] = None,
            downloads_dir: Optional[str] = None
    ) -> None:
        self._set_dirs(program_data_dir, config_file_path, downloads_dir)

    def _set_dirs(
            self,
            program_data_dir: Optional[str],
            config_file_path: Optional[str],
            download_dir: Optional[str]
    ) -> None:
        # Set up main directory and file paths
        self.PROGRAM_DATA_DIR = self._set_program_data_dir(program_data_dir)
        self.CONFIG_FILE_PATH = self._set_config_file_path(config_file_path)
        self.DOWNLOAD_DIR = self._set_download_dir(download_dir)

    def _set_program_data_dir(self, program_data_dir: Optional[str]) -> str:
        """Sets and/or creates the program data dir accordingly.

        If user specifies a custom path for program data dir, it will be honored.
        Folders, including parent directories will be created if they don't exist.

        If there is no specified custom dir, it will just use the default path
        instead.
        """

        if program_data_dir:
            base_dir_abspath = os.path.abspath(program_data_dir)

            if not os.path.exists(base_dir_abspath):
                os.makedirs(base_dir_abspath, exist_ok=True)

            return base_dir_abspath

        else:
            default_path = os.path.join(user_data_path(), APP_NAME)
            base_dir_abspath = os.path.abspath(default_path)

            if not os.path.exists(base_dir_abspath):
                os.makedirs(base_dir_abspath, exist_ok=True)

            return base_dir_abspath

    def _set_config_file_path(self, config_file_path: Optional[str]) -> str:
        """Sets and/or creates the config file path accordingly.

        If user specified a custom config file location, it will be honored
        as long as it exists. If not, it will raise an error.

        If there is no specified custom config path, it will just
        """

        if config_file_path:
            if os.path.isfile(config_file_path):
                return os.path.abspath(config_file_path)
            else:
                error_msg = messages.config_file_loading_error
                console.print(error_msg)
                logger.error(error_msg)
                exit(1)
        else:
            default_path = os.path.join(self.PROGRAM_DATA_DIR, "tk3u8.conf")

            if not os.path.isfile(default_path):
                with open(default_path, "w") as file:
                    toml.dump(DEFAULT_CONFIG, file)

            return default_path

    def _set_download_dir(self, download_dir: Optional[str]) -> str:
        """Sets and/or creates the download dir accordingly.

        If user specifies a custom path for download dir, it will be honored.
        Folders, including parent directories will be created if they don't exist.

        If there is no specified custom dir, it will just use the default path
        instead.
        """

        if download_dir:
            download_dir_abspath = os.path.abspath(download_dir)

            if not os.path.exists(download_dir_abspath):
                os.makedirs(download_dir_abspath, exist_ok=True)

            return download_dir_abspath
        else:
            default_path = os.path.join(user_downloads_path(), APP_NAME)
            download_dir_abspath = os.path.abspath(default_path)

            if not os.path.exists(download_dir_abspath):
                os.makedirs(download_dir_abspath, exist_ok=True)

            return download_dir_abspath



from abc import ABC, abstractmethod
import json

from bs4 import BeautifulSoup

from tk3u8.constants import LiveStatus, Quality
from tk3u8.exceptions import (
    HLSLinkNotFoundError,
    LiveStatusCodeNotFoundError,
    SigiStateMissingError,
    StreamDataNotFoundError,
    UnknownStatusCodeError,
    WAFChallengeError
)
from tk3u8.messages import messages
from tk3u8.session.request_handler import RequestHandler
import logging


logger = logging.getLogger(__name__)


class Extractor(ABC):
    """
    Abstract base class for extracting streaming data for a given username.
    Subclasses must implement methods to fetch source data and extract stream data.
    """

    def __init__(self, username: str, request_handler: RequestHandler):
        self._request_handler = request_handler
        self._username = username

    @abstractmethod
    def get_source_data(self) -> dict:
        """Fetch the raw source data for the user."""

    @abstractmethod
    def get_stream_data(self, source_data: dict) -> dict:
        """Gets the stream data from the extracted source data."""

    @abstractmethod
    def get_live_status(self, source_data: dict) -> LiveStatus:
        """
        Gets the live status code from the extracted source data, then returns
        a LiveStatus constant.
        """

    def get_stream_links(self, stream_data: dict[str, dict]) -> dict:
        """
        This builds the stream links in dict. The qualities are first constructed
        into a list by getting all the values from Quality enum class except for
        the first one ("original"), as this doesn't match with the quality
        specified from the source ("origin").

        After the stream links have been added to the dict, the key "origin" is
        replaced with "original".
        """
        stream_links = {}
        qualities = [quality.value for quality in list(Quality)[1:]]
        qualities.insert(0, "origin")

        for quality_key in qualities:
            quality_dict = {}

            for vid_format in stream_data.keys():
                try:
                    link = stream_data[vid_format]["data"][quality_key]["main"]["hls"]
                except KeyError:
                    link = None

                quality_dict.update({
                    vid_format: link
                })

            stream_links.update({
                quality_key: quality_dict
            })

        stream_links["original"] = stream_links.pop("origin")

        are_stream_links_empty = self._are_hls_stream_links_empty(stream_links)
        if are_stream_links_empty:
            logger.exception(f"{HLSLinkNotFoundError.__name__}: {HLSLinkNotFoundError(self._username)}")
            raise HLSLinkNotFoundError(self._username)

        logger.debug(messages.retrieved_stream_links.format(
            username=self._username,
            stream_links=json.dumps(stream_links, indent=4, ensure_ascii=False)
        ))

        return stream_links

    def _get_live_status(self, status_code: int) -> LiveStatus:
        if status_code == 1:
            return LiveStatus.PREPARING_TO_GO_LIVE
        elif status_code == 2:
            return LiveStatus.LIVE
        elif status_code == 4:
            return LiveStatus.OFFLINE
        else:
            logger.exception(f"{UnknownStatusCodeError.__name__}: {UnknownStatusCodeError(status_code)}")
            raise UnknownStatusCodeError(status_code)

    def _are_hls_stream_links_empty(self, stream_links: dict[str, dict[str, str | None]]) -> bool:
        """
        Checks whether the all stream links contain similar empty strings.
        If satisfied, this function returns True and vice versa.

        This is important because there is a tendency that stream links can be empty.
        Based on testing, this is more likely to happen if you and the user you are
        trying to download is in different server locations.
        """

        h264_bool: bool = False
        h265_bool: bool = False

        for quality in stream_links:
            links_by_codec = stream_links[quality]

            for codec, link in links_by_codec.items():
                if link != "":
                    continue
                else:
                    if codec == "h264":
                        h264_bool = True
                    elif codec == "h265":
                        h265_bool = True

        return all([h264_bool, h265_bool])


class APIExtractor(Extractor):
    def get_source_data(self) -> dict:
        response = self._request_handler.get_data(f"https://www.tiktok.com/api-live/user/room?aid=1988&sourceType=54&uniqueId={self._username}")

        soup = BeautifulSoup(response.text, "html.parser")
        content = json.loads(soup.text)

        logger.debug(messages.fetched_content.format(
            username=self._username,
            content=content
        ))

        return content

    def get_stream_data(self, source_data: dict) -> dict:
        try:
            stream_data = {
                "h264": json.loads(source_data["data"]["liveRoom"]["streamData"]["pull_data"]["stream_data"]),
                "h265": json.loads(source_data["data"]["liveRoom"]["hevcStreamData"]["pull_data"]["stream_data"])
            }

            logger.debug(messages.extracted_stream_data.format(
                username=self._username,
                stream_data=json.dumps(stream_data, indent=4, ensure_ascii=False)
            ))

            return stream_data
        except KeyError:
            logger.exception(f"{StreamDataNotFoundError.__name__}: {StreamDataNotFoundError(self._username)}")
            raise StreamDataNotFoundError(self._username)

    def get_live_status(self, source_data: dict) -> LiveStatus:
        try:
            status_code = source_data["data"]["user"]["status"]
            logger.debug(messages.extracted_status_code.format(
                username=self._username,
                status_code=status_code
            ))

            return self._get_live_status(status_code)
        except KeyError:
            logger.exception(f"{LiveStatusCodeNotFoundError.__name__}: {LiveStatusCodeNotFoundError(self._username)}")
            raise LiveStatusCodeNotFoundError(self._username)


class WebpageExtractor(Extractor):
    def get_source_data(self) -> dict:
        response = self._request_handler.get_data(f"https://www.tiktok.com/@{self._username}/live")

        if "Please wait..." in response.text:
            logger.exception(f"{WAFChallengeError.__name__}: {WAFChallengeError}")
            raise WAFChallengeError()

        soup = BeautifulSoup(response.text, "html.parser")
        script_tag = soup.find("script", {"id": "SIGI_STATE"})

        if not script_tag:
            logger.exception(f"{SigiStateMissingError.__name__}: {SigiStateMissingError}")
            raise SigiStateMissingError()

        content = json.loads(script_tag.text)

        logger.debug(messages.fetched_content.format(
            username=self._username,
            content=content
        ))

        return content

    def get_stream_data(self, source_data: dict) -> dict:
        try:
            stream_data = json.loads(source_data["LiveRoom"]["liveRoomUserInfo"]["liveRoom"]["streamData"]["pull_data"]["stream_data"])
            stream_data = {
                "h264": json.loads(source_data["LiveRoom"]["liveRoomUserInfo"]["liveRoom"]["streamData"]["pull_data"]["stream_data"]),
                "h265": json.loads(source_data["LiveRoom"]["liveRoomUserInfo"]["liveRoom"]["hevcStreamData"]["pull_data"]["stream_data"])
            }

            logger.debug(messages.extracted_stream_data.format(
                username=self._username,
                stream_data=json.dumps(stream_data, indent=4, ensure_ascii=False)
            ))

            return stream_data
        except KeyError:
            logger.exception(f"{StreamDataNotFoundError.__name__}: {StreamDataNotFoundError(self._username)}")
            raise StreamDataNotFoundError(self._username)

    def get_live_status(self, source_data: dict) -> LiveStatus:
        try:
            status_code = source_data["LiveRoom"]["liveRoomUserInfo"]["user"]["status"]
            logger.debug(messages.extracted_status_code.format(
                username=self._username,
                status_code=status_code
            ))

            return self._get_live_status(status_code)
        except KeyError:
            logger.exception(f"{LiveStatusCodeNotFoundError.__name__}: {LiveStatusCodeNotFoundError(self._username)}")
            raise LiveStatusCodeNotFoundError(self._username)



from datetime import datetime
import logging
import os
import time
from yt_dlp import YoutubeDL
from tk3u8.constants import LiveStatus, OptionKey, StreamLink
from tk3u8.cli.console import console, Live, render_lines
from tk3u8.exceptions import DownloadError, QualityNotAvailableError
from tk3u8.messages import messages
from tk3u8.options_handler import OptionsHandler
from tk3u8.core.stream_metadata_handler import StreamMetadataHandler
from tk3u8.paths_handler import PathsHandler


logger = logging.getLogger(__name__)


class Downloader:
    def __init__(
            self,
            paths_handler: PathsHandler,
            stream_metadata_handler: StreamMetadataHandler,
            options_handler: OptionsHandler
    ) -> None:
        self._paths_handler = paths_handler
        self._options_handler = options_handler
        self._stream_metadata_handler = stream_metadata_handler

    def download(self, quality: str) -> None:
        username = self._stream_metadata_handler.get_username()
        wait_until_live = self._options_handler.get_option_val(OptionKey.WAIT_UNTIL_LIVE)
        live_status = self._stream_metadata_handler.get_live_status()
        force_redownload = self._options_handler.get_option_val(OptionKey.FORCE_REDOWNLOAD)
        redownload_attempted = False
        use_h265 = self._options_handler.get_option_val(OptionKey.USE_H265)

        assert isinstance(username, str)
        assert isinstance(wait_until_live, int)
        assert isinstance(live_status, LiveStatus)
        assert isinstance(force_redownload, bool)
        assert isinstance(use_h265, bool)

        while True:
            if live_status in (LiveStatus.OFFLINE, LiveStatus.PREPARING_TO_GO_LIVE):
                # If the user did not use the 'wait until live' option from
                # either the config file or command-line argument, the program
                # will exit momentarily
                if not wait_until_live:
                    if live_status == LiveStatus.OFFLINE:
                        console.print(messages.user_offline.format(username=username))
                        exit(0)
                    elif live_status == LiveStatus.PREPARING_TO_GO_LIVE:
                        console.print(messages.preparing_to_go_live.format(username=username))
                        exit(0)

                # Otherwise, the program will attempt to check the live
                # status every n seconds until the user goes live.
                offline_msg = messages.awaiting_to_go_live.format(username=username)
                self._wait_until_live(offline_msg, live_status)

            if not redownload_attempted:
                console.print(messages.user_is_now_live.format(username=username))
            else:
                console.print(messages.reattempting_download.format(username=username))

            stream_link = self._stream_metadata_handler.get_stream_link(quality, use_h265)
            if not self._is_stream_link_available(stream_link):
                console.print(messages.quality_not_available.format(quality=quality))
                logger.error(f"{QualityNotAvailableError.__name__}: {QualityNotAvailableError()}")
                exit(0)

            self._start_download(username, stream_link)

            if not force_redownload:
                break

            self._show_redownloading_notice()

            self._update_data()
            live_status = live_status = self._stream_metadata_handler.get_live_status()
            redownload_attempted = True

    def _start_download(self, username: str, stream_link: StreamLink) -> None:
        starting_download_msg = messages.starting_download.format(
            username=username,
            stream_link=stream_link
        )
        console.print(starting_download_msg, end="\n\n")
        logger.debug(starting_download_msg)

        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        filename = f"{username}-{timestamp}-{stream_link.quality}"
        filename_with_download_dir = os.path.join(self._paths_handler.DOWNLOAD_DIR, f"{username}", f"{filename}.%(ext)s")

        ydl_opts = {
            'outtmpl': filename_with_download_dir,
            'quiet': False,  # Set to True to suppress output if needed
        }

        try:
            with YoutubeDL(ydl_opts) as ydl:  # type: ignore[arg-type]
                ydl.download([stream_link.link])

                finished_downloading_msg = messages.finished_downloading.format(
                    filename=filename,
                    filename_with_download_dir=filename_with_download_dir.replace('%(ext)s', 'mp4'),
                )
                console.print("\n" + finished_downloading_msg)
                logger.debug(finished_downloading_msg)
        except Exception as e:
            logger.exception(f"{DownloadError.__name__}: {DownloadError(e)}")
            raise DownloadError(e)

    def _is_stream_link_available(self, stream_link: StreamLink) -> bool:
        if stream_link.link is None:
            return False
        return True

    def _wait_until_live(self, offline_msg: str, live_status: LiveStatus) -> None:
        """
        Waits until the stream becomes live, periodically checking the live
        status. This method displays a live-updating message while the stream
        is offline, repeatedly checking the stream's status until it becomes
        live.
        """

        with Live(render_lines(offline_msg)) as live:
            try:
                while not live_status == LiveStatus.LIVE:
                    self._pause_rechecking(live, offline_msg)
                    self._update_data()
                    live_status = self._stream_metadata_handler.get_live_status()
                live.update(render_lines())
            except KeyboardInterrupt:
                live.update(render_lines(offline_msg, messages.cancelled_checking_live))
                exit(0)

    def _update_data(self) -> None:
        self._stream_metadata_handler.update_data()

    def _pause_rechecking(self, live: Live, offline_msg: str) -> None:
        """
        Handles temporarily pausing before rechecking live status,
        and prints the seconds remaining before the next check.
        """

        seconds_left = self._options_handler.get_option_val(OptionKey.TIMEOUT)
        assert isinstance(seconds_left, int)

        seconds_left_len = len(str(seconds_left))
        seconds_extra_space = " " * seconds_left_len

        for remaining in range(seconds_left, -1, -1):
            live.update(render_lines(offline_msg, messages.retrying_to_check_live.format(
                remaining=remaining,
                seconds_extra_space=seconds_extra_space
            )))
            time.sleep(1)

        live.update(render_lines(offline_msg, messages.ongoing_checking_live))

    def _show_redownloading_notice(self) -> None:
        """
        Displays a message notifying the user that they have 5 seconds to
        exit the program to avoid redownloading the stream.

        This only occurs if the user uses the 'FORCE_REDOWNLOAD' option from
        the config file or command-line argument.
        """

        with Live() as live:
            try:
                for remaining_seconds in range(5, -1, -1):
                    live.update(render_lines("\n" + messages.redownloading_notice.format(remaining=remaining_seconds)))
                    time.sleep(1)
            except KeyboardInterrupt:
                live.update(render_lines(messages.exiting_download_reattempt))
                exit(0)



import argparse
from rich_argparse import RichHelpFormatter
from tk3u8.cli.utils import display_version
from tk3u8.constants import Quality


class ArgsHandler():
    _instance = None

    def __new__(cls) -> 'ArgsHandler':
        if cls._instance is None:
            cls._instance = super(ArgsHandler, cls).__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        self._parser: argparse.ArgumentParser = argparse.ArgumentParser(
            description="tk3u8 - A TikTok live downloader",
            formatter_class=RichHelpFormatter
        )
        self._init_args()

    def parse_args(self) -> argparse.Namespace:
        args = self._parser.parse_args()
        return args

    def _init_args(self) -> None:
        self._parser.add_argument(
            "username",
            help="The username to be used for recording live stream",
        )
        self._parser.add_argument(
            "-q",
            choices=[quality.value for quality in Quality],
            default=Quality.ORIGINAL.value,
            dest="quality",
            help="Specify the quality of the video to download. Default: original"
        )
        self._parser.add_argument(
            "--proxy",
            help="The proxy server to use for downloading. Sample format: 127.0.0.1:8080"
        )
        self._parser.add_argument(
            "--wait-until-live",
            action="store_true",
            help="Let the program wait until the user goes live to start downloading stream",
            default=None
        )
        self._parser.add_argument(
            "--timeout",
            help="Set the timeout in seconds before rechecking if the user is live.",
            type=int,
        )
        self._parser.add_argument(
            "--force-redownload",
            action="store_true",
            help="Forces the program to redownload again while user is still live.",
            default=None
        )
        self._parser.add_argument(
            "--use-h265",
            action="store_true",
            help="Use the H.265 (HEVC) encoded live stream instead of H.264 (AVC)",
            default=None
        )
        self._parser.add_argument(
            "--config-file",
            help="The path of the config file to use",
            default=None
        )
        self._parser.add_argument(
            "--download-dir",
            help="The directory where stream downloads will be stored",
            default=None
        )
        self._parser.add_argument(
            "--log-level",
            help="Set the logging level (default: no logging if not used)",
            choices=["DEBUG", "ERROR"],
            dest="log_level"
        )
        self._parser.add_argument(
            "-v", "--version",
            help="Show the program's version",
            action="version",
            version=display_version()
        )



import logging
from typing import List, Optional
from tk3u8.constants import LiveStatus, StreamLink
from tk3u8.cli.console import console
from tk3u8.core.extractor import APIExtractor, Extractor, WebpageExtractor
from tk3u8.core.helper import is_user_exists, is_username_valid
from tk3u8.exceptions import (
    HLSLinkNotFoundError,
    HLSLinkTemporarilyUnavailableError,
    InvalidQualityError,
    InvalidUsernameError,
    NoUsernameEnteredError,
    QualityNotAvailableError,
    SigiStateMissingError,
    StreamDataNotFoundError,
    UserNotFoundError,
    WAFChallengeError
)
from tk3u8.messages import messages
from tk3u8.options_handler import OptionsHandler
from tk3u8.session.request_handler import RequestHandler


logger = logging.getLogger(__name__)


class StreamMetadataHandler:
    """
    Handles the retrieval, validation, and management of stream metadata for a given user.
    This class coordinates the extraction of stream data from multiple sources (API, webpage, etc.),
    validates user input, manages live status, and provides access to stream links and metadata.
    It uses a list of extractor classes to attempt data retrieval in sequence, handling errors and
    falling back as needed.

    Attributes:
        _request_handler (RequestHandler): Handles HTTP requests for data extraction.
        _options_handler (OptionsHandler): Manages configuration options.
        _extractor_classes (List[type[Extractor]]): List of extractor classes to use for data retrieval.
        _source_data (dict): Raw data obtained from the extractor.
        _stream_data (dict): Processed stream data.
        _stream_links (dict): Available stream links by quality.
        _live_status (LiveStatus | None): Current live status of the stream.
        _username (str | None): Username for which metadata is being handled.
    """
    def __init__(self, request_handler: RequestHandler, options_handler: OptionsHandler):
        self._request_handler = request_handler
        self._options_handler = options_handler
        self._extractor_classes: List[type[Extractor]] = [APIExtractor, WebpageExtractor]
        self._source_data: dict = {}
        self._stream_data: dict = {}
        self._stream_links: dict = {}
        self._live_status: LiveStatus | None = None
        self._username: str | None = None

    def initialize_data(self, username: str) -> None:
        with console.status(messages.processing_data):
            self._process_data(username)

    def update_data(self) -> None:
        with console.status(messages.processing_data):
            self._process_data()

    def get_username(self) -> str:
        assert isinstance(self._username, str)

        return self._username

    def get_live_status(self) -> LiveStatus:
        assert isinstance(self._live_status, LiveStatus)

        return self._live_status

    def get_stream_link(self, quality: str, use_h265: bool) -> StreamLink:
        try:
            if quality in self._stream_links:
                assert isinstance(self._username, str)
                codec, formatted_codec_str = ("h265", "H.265") if use_h265 else ("h264", "H.264")
                stream_link = self._stream_links[quality][codec]

                if stream_link == "":
                    logger.exception(f"{HLSLinkTemporarilyUnavailableError.__name__}: {HLSLinkTemporarilyUnavailableError()}")
                    console.print(messages.empty_stream_link_error)
                    exit(0)

                stream_link_obj = StreamLink(quality, stream_link)
                logger.debug(f"Chosen stream link: {stream_link_obj} ({formatted_codec_str})")

                return stream_link_obj

            logger.exception(f"{InvalidQualityError.__name__}: {InvalidQualityError}")
            raise InvalidQualityError()
        except AttributeError:
            logger.exception(f"{QualityNotAvailableError.__name__}: {QualityNotAvailableError}")
            raise QualityNotAvailableError()

    def _process_data(self, username: Optional[str] = None) -> None:
        """
        Processes stream metadata for the given username.

        Whenever the first extractor fails due to the given exceptions, the next
        available extractor will be used. When all of the available extractors
        failed, the program will exit.
        """
        if username:
            self._username = self._validate_username(username)

        assert isinstance(self._username, str)
        logger.debug(messages.processing_data_for_user.format(username=self._username))

        for idx, extractor_class in enumerate(self._extractor_classes):
            logger.debug(messages.trying_extractor.format(
                pos=idx + 1,
                extractor_class_name=extractor_class.__name__
            ))

            try:
                extractor = extractor_class(self._username, self._request_handler)

                self._source_data = self._get_and_validate_source_data(extractor, extractor_class)
                self._live_status = extractor.get_live_status(self._source_data)

                if self._live_status in (LiveStatus.OFFLINE, LiveStatus.PREPARING_TO_GO_LIVE):
                    break

                self._stream_data = extractor.get_stream_data(self._source_data)
                self._stream_links = extractor.get_stream_links(self._stream_data)

                break
            except (
                WAFChallengeError,
                SigiStateMissingError,
                StreamDataNotFoundError,
                HLSLinkNotFoundError
            ) as e:
                if idx != len(self._extractor_classes) - 1:
                    error_msg = messages.current_extractor_failed.format(
                        current_extr_pos=idx + 1,
                        current_extr_name=extractor.__class__.__name__,
                        exc_name=type(e).__name__,
                        next_extr_pos=idx + 2
                    )
                    console.print(error_msg)
                    logger.error(error_msg)
                else:
                    error_msg = messages.last_extractor_failed.format(
                        current_extr_pos=idx + 1,
                        current_extr_name=extractor.__class__.__name__,
                        exc_name=type(e).__name__,
                    )
                    console.print(error_msg)
                    logger.error(error_msg)
                    exit()

    def _validate_username(self, username: str) -> str:
        if not username:
            logger.exception(f"{NoUsernameEnteredError.__name__}: {NoUsernameEnteredError()}")
            console.print(messages.no_username_entered)
            exit(0)

        if not is_username_valid(username):
            logger.exception(f"{InvalidUsernameError.__name__}: {InvalidUsernameError(username)}")
            console.print(messages.invalid_username.format(username=username))
            exit(0)

        logger.debug(f"Entered username: {username}")

        return username

    def _get_and_validate_source_data(self, extractor: Extractor, extractor_class: type[Extractor]) -> dict:
        source_data: dict = extractor.get_source_data()

        assert isinstance(self._username, str)

        if not is_user_exists(extractor_class, source_data):
            logger.exception(f"{UserNotFoundError.__name__}: {UserNotFoundError(self._username)}")
            console.print(messages.account_not_found.format(username=self._username))
            exit(0)

        return source_data



import logging
from typing import Optional
import toml
from toml import TomlDecodeError
from tk3u8.cli.console import console
from tk3u8.constants import OptionKey
from tk3u8.messages import messages
from tk3u8.paths_handler import PathsHandler


OPTION_KEY_DEFAULT_VALUES = {
    OptionKey.SESSIONID_SS: None,
    OptionKey.TT_TARGET_IDC: None,
    OptionKey.PROXY: None,
    OptionKey.WAIT_UNTIL_LIVE: False,
    OptionKey.TIMEOUT: 30,
    OptionKey.FORCE_REDOWNLOAD: False,
    OptionKey.USE_H265: False
}

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class OptionsHandler:
    """
    Responsible for managing application options from multiple sources,
    including command-line arguments and configuration files. It provides
    methods to retrieve option values with a defined precedence
    (command-line arguments > config file > defaults), save argument
    values, and validate configuration files.

    Attributes:
        _paths_initializer (PathsHandler): Helper for resolving
            configuration file paths.
        _args_values (dict): Stores option values provided from the
            command-line arguments.
        _config_values (dict): Stores option values loaded from the configuration file.
    """

    def __init__(self, paths_handler: PathsHandler) -> None:
        self._paths_handler = paths_handler
        self._args_values: dict = {}
        self._config_values: dict = self._load_config_values()

    def get_option_val(self, key: OptionKey) -> Optional[str | int | bool]:
        """
        Retrieves the value for a given option key, checking arguments first, then config file,
        and finally falling back to default values.
        """
        val = self._args_values.get(key.value)
        if val is not None:
            return val

        val = self._config_values.get(key.value)
        if val is not None:
            return val

        return OPTION_KEY_DEFAULT_VALUES.get(key)

    def save_args_values(self, *args: dict, **kwargs: Optional[str | int]) -> None:
        """
        Saves provided argument values into the 'self._args_values',
        accepting both dictionaries and keyword arguments.
        """

        for arg in args:
            if isinstance(arg, dict):
                self._args_values.update(arg)
            else:
                raise TypeError(f"Argument {arg} is not a dict.")

        for key, value in kwargs.items():
            for option_key in list(OptionKey):
                if key in option_key.value:
                    self._args_values.update({key: value})

    def _load_config_values(self) -> dict:
        try:
            with open(self._paths_handler.CONFIG_FILE_PATH, 'r') as file:
                config = self._validate_and_retouch_config(toml.load(file))
                return config
        except (FileNotFoundError, UnicodeDecodeError):
            console.print(messages.config_file_parsing_error)
            logger.error(messages.config_file_parsing_error)
            exit(1)

        except TomlDecodeError as e:
            exc_msg = f'{e.msg} (line {e.lineno} column {e.colno} char {e.pos})'
            formatted_msg = messages.config_file_decoding_error.format(exc_msg=exc_msg)

            console.print(formatted_msg)
            logger.debug(formatted_msg)

            exit(1)

    def _validate_and_retouch_config(self, config: dict) -> dict:
        """Validates the configuration dictionary, ensuring all keys are valid
        option keys andconverting empty string values to None."""

        raw_config: dict = config['config']
        option_keys = [option_key.value for option_key in OptionKey]

        for key, value in raw_config.items():
            if key not in option_keys:
                msg = messages.invalid_option_key.format(key=key)
                console.print(msg)
                logger.debug(msg)

                exit(1)

            if value == "":
                raw_config[key] = None

        return raw_config



from rich.console import Console
from rich.live import Live  # noqa: F401
from rich.table import Table


console = Console()


def render_lines(*args: str) -> Table:
    table = Table.grid()
    for message in args:
        table.add_row(message)
    return table



import logging
from typing import Optional
from tk3u8.cli.console import console
from tk3u8.constants import OptionKey, Quality
from tk3u8.core.downloader import Downloader
from tk3u8.core.stream_metadata_handler import StreamMetadataHandler
from tk3u8.messages import messages
from tk3u8.options_handler import OptionsHandler
from tk3u8.paths_handler import PathsHandler
from tk3u8.session.request_handler import RequestHandler


logger = logging.getLogger(__name__)


class Tk3u8:
    """
    Serves as the main entry point and public API, organizing all core modules
    and functionalities in this single interface, allowing users to easily
    setup their scripts.

    This class is designed to simplify usage by encapsulating the
    initialization and coordination of various internal components,
    such as path handling, options management, HTTP requests, stream
    metadata processing, and downloading logic. Users are encouraged to
    interact with this class directly when integrating tk3u8 into their
    scripts.
    """
    def __init__(
            self,
            program_data_dir: Optional[str] = None,
            config_file_path: Optional[str] = None,
            downloads_dir: Optional[str] = None
    ) -> None:
        logger.debug("Initializing Tk3u8 class")
        self._paths_handler = PathsHandler(program_data_dir, config_file_path, downloads_dir)
        self._options_handler = OptionsHandler(self._paths_handler)
        self._request_handler = RequestHandler(self._options_handler)
        self._stream_metadata_handler = StreamMetadataHandler(
            self._request_handler,
            self._options_handler
        )
        self._downloader = Downloader(
            self._paths_handler,
            self._stream_metadata_handler,
            self._options_handler
        )

    def download(
            self,
            username: str,
            quality: str = Quality.ORIGINAL.value,
            wait_until_live: Optional[bool] = None,
            timeout: Optional[int] = None,
            force_redownload: Optional[bool] = None,
            use_h265: Optional[bool] = None
    ) -> None:
        """
        Downloads a stream for the specified user with the given quality and options.
        Args:
            username (str): The username of the stream to download.
            quality (str, optional): The desired stream quality. Defaults to
                original".
            wait_until_live (bool, optional): Whether to wait until the stream
                is live before downloading. Defaults to False.
            timeout (int, optional): The timeout (in seconds) before rechecking
                if the user is llive. Defaults to 30.
            force_redownload (bool, optional): Force re-download while the user
                is live. Use this if you encounter auto-stopping of download.
                Defaults to False.
        """
        self._options_handler.save_args_values(
            wait_until_live=wait_until_live,
            timeout=timeout,
            force_redownload=force_redownload,
            use_h265=use_h265
        )
        self._stream_metadata_handler.initialize_data(username)
        self._downloader.download(quality)

    def set_proxy(self, proxy: str | None) -> None:
        """
        Sets the proxy configuration.

        Args:
            proxy (str | None): The proxy address to set (e.g., 127.0.0.1:8080).
        """
        self._options_handler.save_args_values({OptionKey.PROXY.value: proxy})

        new_proxy = self._options_handler.get_option_val(OptionKey.PROXY)
        assert isinstance(new_proxy, (str, type(None)))

        self._request_handler.update_proxy(new_proxy)

    def set_cookies(self, cookies: dict) -> None:
        for key, value in cookies.items():
            if key == OptionKey.SESSIONID_SS.value:
                self._options_handler.save_args_values({OptionKey.SESSIONID_SS.value: value})
            elif key == OptionKey.TT_TARGET_IDC.value:
                self._options_handler.save_args_values({OptionKey.TT_TARGET_IDC.value: value})
            else:
                error_msg = messages.invalid_cookie_key_error.format(key=key)
                console.print(error_msg)
                logger.error(error_msg)
                exit(1)

        new_sessionid_ss = self._options_handler.get_option_val(OptionKey.SESSIONID_SS)
        new_tt_target_idc = self._options_handler.get_option_val(OptionKey.TT_TARGET_IDC)

        assert isinstance(new_sessionid_ss, str)
        assert isinstance(new_tt_target_idc, str)

        self._request_handler.update_cookies(new_sessionid_ss, new_tt_target_idc)



import logging
import random
import requests
from requests.exceptions import ConnectionError, ReadTimeout
from tk3u8.constants import USER_AGENT_LIST, OptionKey
from tk3u8.exceptions import RequestFailedError
from tk3u8.options_handler import OptionsHandler


logger = logging.getLogger(__name__)


class RequestHandler:
    """
    Handles HTTP requests with session management, proxy, and cookie support.
    This class manages a requests.Session object, allowing for persistent cookies,
    proxy configuration, and dynamic User-Agent selection. It provides methods to
    fetch data from URLs, update proxy settings, and configure session cookies
    based on provided options.

    Args:
        options_handler (OptionsHandler): An instance responsible for providing
            configuration options such as cookies and proxy settings.

    Attributes:
        _options_handler (OptionsHandler): Stores the options handler instance.
        _session (requests.Session): The session object used for HTTP requests.
        _response (requests.Response): Stores the most recent response object.
    """
    def __init__(self, options_handler: OptionsHandler) -> None:
        self._options_handler = options_handler
        self._session: requests.Session
        self._initialize_session()

    def get_data(self, url: str) -> requests.Response:
        retries = 3
        exc_msg: str = ""

        for retry in range(1, retries + 1):
            try:
                response = self._session.get(url)
                status_code = response.status_code

                if status_code != 200:
                    raw_exc_msg = f"Request error due to status code: {status_code}"
                    exc_msg = f"{RequestFailedError.__name__}: {RequestFailedError(raw_exc_msg)}"

                    if retry == 3:
                        break

                    logger.warning(exc_msg)
                    continue

                return response

            except (ConnectionError, ReadTimeout) as e:
                logger.warning(f"{type(e).__name__} occurred on attempt #{retry}: {e}")

                if retry == 3:
                    exc_msg = f"{RequestFailedError.__name__}: {RequestFailedError(str(e))}"
                    logger.exception(exc_msg)

                    break

                # Re-initalizes the sesssion instance to discard the previous
                # instance as the established connection of it is likely to
                # be staled, so we need to re-initialized this to create new
                # connection.
                self._initialize_session(reinitialized=True)

        raise RequestFailedError(exc_msg)

    def update_proxy(self, proxy: str | None) -> None:
        if proxy:
            self._session.proxies.update({
                    "http": proxy,
                    "https": proxy
            })

            logger.debug(f"Proxy updated to: {proxy}")

    def update_cookies(self, sessionid_ss: str, tt_target_idc: str) -> None:
        self._session.cookies.update({
                "sessionid_ss": sessionid_ss,
                "tt-target-idc": tt_target_idc
            })
        logger.debug(f"'sessionid_ss' cookie updated to: {sessionid_ss}")
        logger.debug(f"'sessionid_ss' cookie updated to: {tt_target_idc}")

    def _initialize_session(self, reinitialized: bool = False) -> None:
        if hasattr(self, '_session') and self._session:
            try:
                self._session.close()
                logger.debug("Previous requests' session closed.")
            except Exception as e:
                logger.warning(f"Error closing previous requests' session: {e}")

        if reinitialized:
            logger.debug("Re-initializing requests.Session")

        self._session = requests.Session()
        self._setup_cookies()
        self._setup_proxy()
        self._session.headers.update({
            "User-Agent": self._get_random_user_agent()
        })

        logger.debug("New requests.Session initialized.")

    def _setup_cookies(self) -> None:
        sessionid_ss = self._options_handler.get_option_val(OptionKey.SESSIONID_SS)
        tt_target_idc = self._options_handler.get_option_val(OptionKey.TT_TARGET_IDC)

        assert isinstance(sessionid_ss, (str, type(None)))
        assert isinstance(tt_target_idc, (str, type(None)))

        if sessionid_ss is None and tt_target_idc is None:
            return

        if sessionid_ss:
            self._session.cookies.update({
                "sessionid_ss": sessionid_ss
            })
            logger.debug(f"'sessionid_ss' cookie set to: {sessionid_ss}")

        if tt_target_idc:
            self._session.cookies.update({
                "tt-target-idc": tt_target_idc
            })
            logger.debug(f"'tt-target-idc' cookie set to: {tt_target_idc}")

    def _setup_proxy(self) -> None:
        proxy = self._options_handler.get_option_val(OptionKey.PROXY)
        assert isinstance(proxy, (str, type(None)))

        if proxy:
            self.update_proxy(proxy)

    def _get_random_user_agent(self) -> str:
        random_ua = random.choice(USER_AGENT_LIST)

        logger.debug(f"Random User-Agent selected: {random_ua}")
        return random_ua



import re
from tk3u8.core.extractor import APIExtractor, Extractor, WebpageExtractor
from tk3u8.exceptions import InvalidExtractorError


def is_username_valid(username: str) -> bool:
    pattern = r"^[a-z0-9_.]{2,24}$"
    match = re.match(pattern, username)

    if match:
        return True
    return False


def is_user_exists(extractor: type[Extractor], source_data: dict) -> bool:
    if extractor == WebpageExtractor:
        if source_data.get("LiveRoom"):
            return True
        return False

    elif extractor == APIExtractor:
        message = source_data["message"]
        if message == "user_not_found":
            return False
        return True

    else:
        raise InvalidExtractorError()



from tk3u8.constants import OptionKey


class RequestFailedError(Exception):
    """Custom exception for failed HTTP requests.

    Raised when an HTTP request fails due to network issues,
    invalid responses, or other related errors.
    """

    def __init__(self, exc_msg: str) -> None:
        self.message = f"Request failed: {exc_msg})"
        super().__init__(self.message)


class WAFChallengeError(Exception):
    """Custom exception when 'Please wait...' message appears when extracting data
    from source."""

    def __init__(self) -> None:
        self.message = "Source extraction failed due to WAF. Please try again later."
        super().__init__(self.message)


class SigiStateMissingError(Exception):
    """Custom exception for failed data extraction from the SIGI_STATE script tag.

    Raised when the SIGI_STATE script tag isn't found from the webpage.
    """

    def __init__(self) -> None:
        self.message = "SIGI_STATE script not found."
        super().__init__(self.message)


class UserNotLiveError(Exception):
    """Custom exception when user is not live."""

    def __init__(self, username: str) -> None:
        self.message = f"User @{username} is not live."
        super().__init__(self.message)


class UserNotFoundError(Exception):
    """Custom exception whenever data extraction from specified user fails.

    This exception will be raised whenever the account is private or the
    account doesn't exist.
    """

    def __init__(self, username: str) -> None:
        self.message = f"The user account @{username} is likely a private account, or it doesn't exist at all."
        super().__init__(self.message)


class InvalidUsernameError(Exception):
    """Custom exception whenever username entered is invalid."""

    def __init__(self, username: str) -> None:
        self.message = (
            f"The username @{username} is invalid. Ensure the username is at least "
            "2 characters long, contains only lowercase letters, numbers, underscores, "
            "and/or periods, and is up to 24 characters in length."
        )
        super().__init__(self.message)


class NoUsernameEnteredError(Exception):
    """Custom exception when no username is entered."""

    def __init__(self) -> None:
        self.message = "No username was entered. Please provide a valid username."
        super().__init__(self.message)


class UserPreparingForLiveError(Exception):
    """Custom exception when the user is preparing to go live.
    """

    def __init__(self, username: str) -> None:
        self.message = f"User @{username} is preparing to go live. Try again in a minute or two to be able to download the stream."
        super().__init__(self.message)


class UnknownStatusCodeError(Exception):
    """Custom exception whenever the status code returned isn't 2 or 4.

    This exception is a weird one, but I still implemented in case that there might be some
    situation that the status integer returns beside 2 or 4 from
    ["LiveRoom"]["liveRoomUserInfo"]["user"]["status"]. This can be useful
    for debugging if TikTok made some changes in their end.
    """

    def __init__(self, status_code: int) -> None:
        self.message = f"Invalid status code. (Status code: {status_code} {type(status_code)})"
        super().__init__(self.message)


class InvalidQualityError(Exception):
    """Custom exception when quality arg is incorrectly entered."""

    def __init__(self) -> None:
        self.message = "Invalid video quality entered. Supported args: [-q {original,uhd,hd,ld,sd}])"
        super().__init__(self.message)


class QualityNotAvailableError(Exception):
    """Custom exception when quality is not available for download."""

    def __init__(self) -> None:
        self.message = "The requested video quality is not available for download."
        super().__init__(self.message)


class LinkNotAvailableError(Exception):
    """Custom exception when the stream link isn't available for some reason."""

    def __init__(self) -> None:
        self.message = "Stream link can't be retrieved. Please try again."
        super().__init__(self.message)


class StreamDataNotFoundError(Exception):
    """Custom exception when the stream data can't be scraped."""

    def __init__(self, username: str) -> None:
        self.message = f"Stream data can't be retrieved from user @{username}. Please try again."
        super().__init__(self.message)


class LiveStatusCodeNotFoundError(Exception):
    """Custom exception when the live status code failed to be retrieved."""

    def __init__(self, username: str) -> None:
        self.message = f"Live status code could not be retrieved from user @{username}."
        super().__init__(self.message)


class HLSLinkNotFoundError(Exception):
    """Custom exception when the HLS stream link isn't available, even though
    there is a stream going on."""

    def __init__(self, username: str) -> None:
        self.message = f"HLS stream link not found for user @{username}."
        super().__init__(self.message)


class HLSLinkTemporarilyUnavailableError(Exception):
    """Custom exception when the HLS stream link isn't available temporarily
    for some reason.

    The difference with the exception 'HLSLinkNotFoundError' is that this exception
    will be used if the HLS link becomes unavaiable initially, but becomes available
    again whenever user retries to download the live stream. The other exception is
    only used if there is no actually available HLS stream links persistently.
    """

    def __init__(self) -> None:
        self.message = "Cannot proceed with downloading as the stream link was somehow unavailable during stream data extraction. Try downloading again."
        super().__init__(self.message)


class InvalidArgKeyError(Exception):
    """Custom exception when an invalid key is encountered."""

    def __init__(self, key: OptionKey) -> None:
        self.message = f"The key '{key}' is invalid or not recognized."
        super().__init__(self.message)


class FileParsingError(Exception):
    """Custom exception for when there is a problem parsing the file."""

    def __init__(self) -> None:
        self.message = "Error parsing config file."
        super().__init__(self.message)


class InvalidCookieError(Exception):
    """Custom exception when user improperly sets cookie in the config file."""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class DownloadError(Exception):
    """Custom exception when there is an issue downloading with yt-dlp."""

    def __init__(self, e: Exception) -> None:
        super().__init__(f"Download failed with error: {e}")


class InvalidExtractorError(Exception):
    """Custom exception raised when an invalid extractor is used."""

    def __init__(self) -> None:
        super().__init__("The specified extractor is invalid or has failed.")



from tk3u8.cli.args_handler import ArgsHandler
from tk3u8.cli.logging import setup_logging


def start_cli() -> None:
    from tk3u8.core.model import Tk3u8

    ah = ArgsHandler()
    args = ah.parse_args()

    username = args.username
    quality = args.quality
    proxy = args.proxy
    wait_until_live = args.wait_until_live
    timeout = args.timeout
    log_level = args.log_level
    force_redownload = args.force_redownload
    use_h265 = args.use_h265
    config_file_path = args.config_file
    download_dir = args.download_dir

    setup_logging(log_level)

    tk3u8 = Tk3u8(config_file_path=config_file_path, downloads_dir=download_dir)
    tk3u8.set_proxy(proxy)
    tk3u8.download(
        username=username,
        quality=quality,
        wait_until_live=wait_until_live,
        timeout=timeout,
        force_redownload=force_redownload,
        use_h265=use_h265
    )



from dataclasses import dataclass
from enum import Enum


@dataclass
class StreamLink:
    quality: str
    link: str


class StatusCode(Enum):
    OK = 200
    BAD_REQUEST = 400
    NOT_FOUND = 404
    FORBIDDEN = 403
    UNAUTHORIZED = 401
    SERVER_ERROR = 500
    GATEWAY_TIMEOUT = 504
    SERVICE_UNAVAILABLE = 503


class Quality(Enum):
    ORIGINAL = "original"
    UHD_60 = "uhd_60"
    UHD = "uhd"
    HD_60 = "hd_60"
    HD = "hd"
    LD = "ld"
    SD = "sd"


class LiveStatus(Enum):
    LIVE = "live"
    PREPARING_TO_GO_LIVE = "preparting_to_go_live"
    OFFLINE = "offline"


class OptionKey(Enum):
    SESSIONID_SS = "sessionid_ss"
    TT_TARGET_IDC = "tt_target_idc"
    PROXY = "proxy"
    USERNAME = "username"
    QUALITY = "quality"
    WAIT_UNTIL_LIVE = "wait_until_live"
    TIMEOUT = "timeout"
    FORCE_REDOWNLOAD = "force_redownload"
    USE_H265 = "use_h265"


@dataclass
class Messages:
    """A collection of message templates used throughout the program."""
    user_offline: str = "User [b]@{username}[/b] is [red]currently offline[/red]."
    preparing_to_go_live: str = "User [b]@{username}[/b] is preparing to go live. Try again in a minute or two to be able to download the stream."
    user_is_now_live: str = "User [b]@{username}[/b] is now [b][green]streaming live[/b][/green]."
    reattempting_download: str = "[grey50]Reattempting download for user [b]@{username}[/b]...[/grey50]"
    awaiting_to_go_live: str = "User [b]@{username}[/b] is [red]currently offline[/red]. Awaiting [b]@{username}[/b] to start streaming..."
    quality_not_available: str = "[grey50]Cannot proceed with downloading. The chosen quality [b]({quality})[/b] is not available for download.[/grey50]"
    empty_stream_link_error: str = "Cannot proceed with downloading as the stream link was somehow unavailable during stream data extraction. Try downloading again."
    starting_download: str = "Starting download for user [b]@{username}[/b] [grey50](quality: {stream_link.quality}, stream Link: {stream_link.link})[/grey50]"
    finished_downloading: str = "[green]Finished downloading[/green] [b]{filename}.mp4[/b] [grey50](saved at: {filename_with_download_dir})[/grey50]"
    cancelled_checking_live: str = "[grey50]Checking cancelled by user. Exiting...[/grey50]"
    retrying_to_check_live: str = "[bold yellow]Retrying in {remaining} seconds{seconds_extra_space}"
    ongoing_checking_live: str = "[grey50]Checking...[/grey50]"
    redownloading_notice: str = "You are using the [yellow]FORCE_REDOWNLOAD[/yellow] option. You have {remaining} seconds to exit the program, or the program will attempt to redownload again."
    processing_data: str = "Processing data..."
    processing_data_for_user: str = "Processing data for user @{username}"
    trying_extractor: str = "Trying extractor #{pos}: {extractor_class_name}"
    current_extractor_failed: str = "[grey50]Extractor #{current_extr_pos} ({current_extr_name}) failed due to [b]{exc_name}[/b]. Trying next extractor method (Extractor #{next_extr_pos})[grey50]"
    last_extractor_failed: str = "[grey50]Extractor #{current_extr_pos} ({current_extr_name}) failed due to [b]{exc_name}[/b]. No more extractors to be used. The program will now exit.[grey50]"
    invalid_username: str = (
        "The username [b]{username}[/b] is [red]invalid[/red]. Ensure the username is at least 2wd "
        "characters long, contains only lowercase letters, numbers, underscores, and/or periods, "
        "and is up to 24 characters in length."
    )
    no_username_entered: str = "No username was entered. Please provide a valid username."
    account_not_found: str = "User [b]@{username}[/b] is likely a private account, or it doesn't exist at all."
    fetched_content: str = "Fetched content for user @{username}: {content}"
    retrieved_stream_links: str = "Retrieved stream links for user @{username}: {stream_links}"
    extracted_stream_data: str = "Extracted stream_data for user @{username}: {stream_data}"
    extracted_status_code: str = "Extracted status_code for user @{username}: {status_code}"
    exiting_download_reattempt: str = "[grey50]Reattempting download cancelled by user. Exiting...[/grey50]"
    invalid_option_key: str = "Option key [b]{key}[/b] is invalid. Please ensure you entered a valid option key in your config file."
    config_file_decoding_error: str = "Error decoding config file due to: '[yellow]{exc_msg}[/yellow]'"
    config_file_loading_error: str = "Config file path is not valid. Ensure that the path is correct and the config file actually exists."
    config_file_parsing_error: str = "Error parsing config file."
    invalid_cookie_key_error: str = "Cookie key '{key}' is invalid. Ensure that the cookie key is either 'sessionid_ss' or 'tt_target_idc'."


APP_NAME = "tk3u8"


# Default configuration settings
DEFAULT_CONFIG = {
    "config": {
        OptionKey.SESSIONID_SS.value: "",
        OptionKey.TT_TARGET_IDC.value: "",
        OptionKey.PROXY.value: ""
    }
}


# Default user agent
USER_AGENT_LIST = [
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/126.0.6478.127 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36 OPR/109.0.0.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 14_4_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36 OPR/109.0.0.0",
    "Mozilla/5.0 (X11; Linux i686; rv:124.0) Gecko/20100101 Firefox/124.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:124.0) Gecko/20100101 Firefox/124.0"
]
